/**
 * Copyright 2006-2007 by jFramework.org.
 */
package org.jframework.core.util;

import junit.framework.TestCase;
import junitx.framework.ArrayAssert;

import java.math.BigInteger;
import java.util.*;

import org.apache.log4j.BasicConfigurator;

/**
 * Tests bean utilities.
 *
 * @author mgridnev
 * @version 2.0, March 1, 2007
 */
public class BeanUtilsTest extends TestCase
{
    protected void setUp() throws Exception
    {
        BasicConfigurator.configure();
    }

    protected void tearDown() throws Exception
    {
        BasicConfigurator.resetConfiguration();
    }

    /**
     * Ensures Class semantics are observed
     */
    public void testNewInstanceSemantics()
    {
        final Object bean = BeanUtils.newInstance(Bean.class);
        final Object bean2 = BeanUtils.newInstance(Bean.class);

        assertTrue(bean.getClass() == bean2.getClass());
    }

    /**
     * Tests instance creation
     */
    public void testNewInstance()
    {
        Object bean = BeanUtils.newInstance(Bean.class);

        assertTrue(bean instanceof Bean);

        final int id = 1;
        final String name = "name";
        bean = BeanUtils.newInstance(Bean.class, id, name);

        assertEquals(id, ((Bean) bean).getID());
        assertEquals(name, ((Bean) bean).getName());

        // Multiple ctors
        assertTrue(BeanUtils.newInstance(String.class) != null);

        // Argument conversion
        bean = BeanUtils.newInstance(StringConversionBean.class, String.class.getName());
        assertTrue(bean instanceof StringConversionBean);
        assertTrue(((StringConversionBean) bean).getType() == String.class);
    }


    /**
     * Tests property name validation
     */
    public void testValidateFullPropertyName()
    {
        String[] properties = {
            "name", "_", "1", "name.name", "name[]", "name[0]", "name[123]", "name[a]", "name[ ]", "name[abc[123]]", "name[key].name[key]"
        };

        for (final String property : properties) {
            assertFullPropertyName(true, property);
        }

        properties = new String[]{
            "", ".", "name.", ".name", "[]", "a[", "a]"
        };

        for (final String property : properties) {
            assertFullPropertyName(false, property);
        }
    }

    /**
     * Asserts full property name
     */
    private void assertFullPropertyName(final boolean valid, final String name)
    {
        try {
            BeanUtils.validatePropertyName(name);
            if (!valid) {
                fail(name + " must fail validation.");
            }
        } catch (RuntimeException e) {
            if (valid) {
                fail(name + " must pass validation.");
            }
        }
    }

    /**
     * Tests property name parser
     */
    public void testPropertyName()
    {
        final String[][] checks = {
            {"name", "name", null},
            {"na_m0", "na_m0", null},
            {"_name", "_name", null},
            {"name_", "name_", null},
            {"name[0]", "name", "0"},
            {"name[123]", "name", "123"},
            {"name[a]", "name", "a"},
            {"name[-1]", "name", "-1"},
            {"name[]", "name", ""},
        };

        for (final String[] check : checks) {
            verifyPropertyName(check[0], check[1], check[2]);
        }

        final String[] properties = {
            "", "na-me", "name[", "name]",
        };

        for (final String property : properties) {
            assertInvalidPropertyName(property);
        }
    }

    /**
     * Verifies property name
     *
     * @param fullName full property name
     * @param baseName base name
     * @param key      property index
     */
    private static void verifyPropertyName(final String fullName, final String baseName, final String key)
    {
        final BeanUtils.PropertyName name = new BeanUtils.PropertyName(fullName);
        assertEquals(baseName, name.getBaseName());
        assertEquals(key, name.getKey());
    }

    /**
     * Asserts invalid property name
     *
     * @param fullName full property name
     */
    private static void assertInvalidPropertyName(final String fullName)
    {
        try {
            new BeanUtils.PropertyName(fullName);
            fail(fullName);
        } catch (RuntimeException e) {
        }
    }

    /**
     * Tests argument conversion
     */
    public void testArgumentConversion()
    {
        final Object[] arguments = {
            "value",
            "123",
            "abc".getBytes(),
            new byte[]{0}
        };

        final Class[] types = {
            String.class,
            byte[].class,
            String.class,
            BigInteger.class
        };

        final Object[] typedArguments = BeanUtils.convertArguments(arguments, types);

        assertEquals(arguments[0], typedArguments[0]);
        ArrayAssert.assertEquals(((String) arguments[1]).getBytes(), (byte[]) typedArguments[1]);
        assertEquals(new String((byte[]) arguments[2]), typedArguments[2]);
        assertEquals(new BigInteger((byte[]) arguments[3]), typedArguments[3]);
    }

    /**
     * Test property type extraction
     */
    public void testGetPropertyType()
    {
        final Bean bean = BeanUtils.newInstance(Bean.class);
        bean.setID(1);

        assertEquals(int.class, BeanUtils.getPropertyType(bean, "ID"));

        assertEquals(String.class, BeanUtils.getPropertyType(bean, "name"));
        assertNull(BeanUtils.getProperty(bean, "name")); // even though the value is null
    }

    /**
     * Tests simple get
     */
    public void testGet()
    {
        final Bean bean = BeanUtils.newInstance(Bean.class);
        bean.setID(1);
        bean.setName("test");

        assertEquals(1, BeanUtils.getProperty(bean, "ID"));
        assertEquals("test", BeanUtils.getProperty(bean, "name"));

        try {
            BeanUtils.getProperty(bean, "foo");
            fail("Invalid property");
        } catch (Exception e) {
        }
    }

    /**
     * Tests simple set
     */
    public void testSet()
    {
        final Bean bean = BeanUtils.newInstance(Bean.class);
        bean.setID(0);
        bean.setName("");

        BeanUtils.setProperty(bean, "ID", 1);
        assertEquals(1, bean.getID());

        BeanUtils.setProperty(bean, "ID", "2");
        assertEquals(2, bean.getID());

        BeanUtils.setProperty(bean, "name", "test");
        assertEquals("test", bean.getName());

        BeanUtils.setProperty(bean, "name", (Object) null);
        assertNull(bean.getName());

        try {
            BeanUtils.setProperty(bean, "foo", 1);
            fail("Invalid property name");
        } catch (RuntimeException e) {
        }
    }

    /**
     * Tests multi setter property
     */
    public void testOverloadedSet()
    {
        final OverloadedSetterBean bean = new OverloadedSetterBean();

        final String strValue = "Value";
        final boolean fValue = true;

        BeanUtils.setProperty(bean, "property", strValue);
        BeanUtils.setProperty(bean, "property", fValue);

        assertEquals(strValue, bean.getString());
        assertEquals(fValue, bean.getBoolean());
    }

    /**
     * Tests setting of a primitive type property
     */
    public void testPrimitiveSet()
    {
        final PrimitiveTypeBean bean = new PrimitiveTypeBean();

        Object[][] properties = {
            {"byte", (byte) 0},
            {"short", (short) 1},
            {"int", 2},
            {"long", (long) 3},
            {"float", 4.0f},
            {"double", 5.0},
            {"char", 'a'},
            {"boolean", true},
        };

        for (Object[] property : properties) {
            BeanUtils.setProperty(bean, (String) property[0], property[1]);
        }

        assertEquals(0, bean.getByte());
        assertEquals(1, bean.getShort());
        assertEquals(2, bean.getInt());
        assertEquals(3, bean.getLong());
        assertTrue(4.0 == bean.getFloat());
        assertTrue(5.0 == bean.getDouble());
        assertEquals('a', bean.getChar());
        assertEquals(true, bean.isBoolean());
    }

    /**
     * Tests setting a property requiring a string conversion to specific argument type
     */
    public void testStringConversionSet() throws ClassNotFoundException
    {
        final StringConversionBean bean = new StringConversionBean();

        // Class
        String value = String.class.getName();
        String property = "type";
        BeanUtils.setProperty(bean, property, value);
        assertEquals(Class.forName(value), BeanUtils.getProperty(bean, property));

        // Char array
        property = "chars";
        BeanUtils.setProperty(bean, property, value);
        assertEquals(value, new String((char[]) BeanUtils.getProperty(bean, property)));

        // Byte array
        property = "bytes";
        BeanUtils.setProperty(bean, property, value);
        assertEquals(value, new String((byte[]) BeanUtils.getProperty(bean, property)));

        // Date
        property = "date";
        value = Long.toString(System.currentTimeMillis());
        BeanUtils.setProperty(bean, property, value);
        assertEquals(value, Long.toString(((Date) BeanUtils.getProperty(bean, property)).getTime()));
    }

    /**
     * Tests setting a property with several arguments
     */
    public void testMultipleArgumentSet()
    {
        final MultiArgSetterBean bean = new MultiArgSetterBean();

        BeanUtils.setProperty(bean, "foo", "bar");
        assertEquals("bar", bean.getFoo());

        BeanUtils.setProperty(bean, "foo", "foo", "bar");
        assertEquals("foobar", bean.getFoo());
    }

    /**
     * Tests indexed getter
     */
    public void testIndexedGet()
    {
        final IndexedPropertyBean bean = new IndexedPropertyBean();
        final String value = "value";
        final String value2 = "value2";

        bean.getValues().add(value);
        bean.getValues().add(value2);

        // Using indexed accessors
        assertEquals(value, BeanUtils.getProperty(bean, "value[0]"));
        assertEquals(value2, BeanUtils.getProperty(bean, "value[1]"));

        // Using list directly
        assertEquals(value, BeanUtils.getProperty(bean, "values[0]"));
        assertEquals(value2, BeanUtils.getProperty(bean, "values[1]"));
    }

    /**
     * Tests indexed setter
     */
    public void testIndexedSet()
    {
        final IndexedPropertyBean bean = new IndexedPropertyBean();
        bean.getValues().add("");
        bean.getValues().add("");

        // Using indexed accessors
        String value = "value";
        BeanUtils.setProperty(bean, "value[0]", value);

        String value2 = "value2";
        BeanUtils.setProperty(bean, "value[1]", value2);

        assertEquals(value, BeanUtils.getProperty(bean, "value[0]"));
        assertEquals(value2, BeanUtils.getProperty(bean, "value[1]"));

        // Using list directly
        value = "value.new";
        BeanUtils.setProperty(bean, "values[0]", value);

        value2 = "value2.new";
        BeanUtils.setProperty(bean, "values[1]", value2);

        assertEquals(value, BeanUtils.getProperty(bean, "values[0]"));
        assertEquals(value2, BeanUtils.getProperty(bean, "values[1]"));
    }

    /**
     * Tests indexed getter
     */
    public void testKeyedGet()
    {
        final KeyedPropertyBean bean = new KeyedPropertyBean();
        final String key = "key";
        final String value = "value";

        bean.getMap().put(key, value);

        // Using map directly
        assertEquals(value, BeanUtils.getProperty(bean, "map[" + key + "]"));
    }

    /**
     * Tests indexed setter
     */
    public void testKeyedSet()
    {
        final KeyedPropertyBean bean = new KeyedPropertyBean();
        final String key = "key";
        final String value = "value";

        bean.getMap().put(key, value);

        // Using map directly
        BeanUtils.setProperty(bean, "map[" + key + "]", value);

        assertEquals(value, BeanUtils.getProperty(bean, "map[" + key + "]"));
    }

    /**
     * Tests setting a hierarchical property, i.e., dotted notation property
     */
    public void testDottedPropertyNotation()
    {
        final ParentBean bean = new ParentBean();
        final Integer id = 10;

        BeanUtils.setProperty(bean, "child.ID", id);
        assertEquals(id, BeanUtils.getProperty(bean, "child.ID"));
        assertEquals(id.intValue(), bean.getChild().getID());

        try {
            BeanUtils.getProperty(bean, "child.");
            fail();
        } catch (RuntimeException e) {
        }

        try {
            BeanUtils.getProperty(bean, ".ID");
            fail();
        } catch (RuntimeException e) {
        }

        try {
            BeanUtils.getProperty(bean, ".");
            fail();
        } catch (RuntimeException e) {
        }
    }

    /**
     * Test parent bean initialization when setting 'parent.child' type property and <tt>getParent()</tt> returns <code>null</code>
     */
    public void testParentPropertyNullSet()
    {
        final ParentBean bean = new ParentBean();
        bean.setChild(null); // null the parent property

        try {
            BeanUtils.getProperty(bean, "child.ID");
            fail();
        } catch (NullPointerException e) {
        }

        final Integer id = 10;

        BeanUtils.setProperty(bean, "child.ID", id);
        assertEquals(id, BeanUtils.getProperty(bean, "child.ID"));
        assertEquals(id.intValue(), bean.getChild().getID());
    }

    /**
     * A bean containing simpe properties
     */
    private static class Bean
    {
        int ID;
        String name;
        boolean flag;
        boolean flag2;

        public Bean()
        {
        }

        public Bean(final int id, final String name)
        {
            this.ID = id;
            this.name = name;
        }

        public int getID()
        {
            return ID;
        }

        public void setID(final int ID)
        {
            this.ID = ID;
        }

        public String getName()
        {
            return name;
        }

        public void setName(final String name)
        {
            this.name = name;
        }

        public boolean isFlag()
        {
            return flag;
        }

        public void setFlag(final boolean value)
        {
            flag = value;
        }

        public boolean getFlag2()
        {
            return flag2;
        }

        public void setFlag2(final boolean value)
        {
            flag2 = value;
        }
    }

    /**
     * A bean containing all primitive types
     */
    private static class PrimitiveTypeBean
    {
        byte aByte;
        short aShort;
        int anInt;
        long aLong;
        float aFloat;
        double aDouble;
        char aChar;
        boolean aBoolean;

        public byte getByte()
        {
            return aByte;
        }

        public void setByte(final byte aByte)
        {
            this.aByte = aByte;
        }

        public short getShort()
        {
            return aShort;
        }

        public void setShort(final short aShort)
        {
            this.aShort = aShort;
        }

        public int getInt()
        {
            return anInt;
        }

        public void setInt(final int anInt)
        {
            this.anInt = anInt;
        }

        public long getLong()
        {
            return aLong;
        }

        public void setLong(final long aLong)
        {
            this.aLong = aLong;
        }

        public float getFloat()
        {
            return aFloat;
        }

        public void setFloat(final float aFloat)
        {
            this.aFloat = aFloat;
        }

        public double getDouble()
        {
            return aDouble;
        }

        public void setDouble(final double aDouble)
        {
            this.aDouble = aDouble;
        }

        public char getChar()
        {
            return aChar;
        }

        public void setChar(final char aChar)
        {
            this.aChar = aChar;
        }

        public boolean isBoolean()
        {
            return aBoolean;
        }

        public void setBoolean(final boolean aBoolean)
        {
            this.aBoolean = aBoolean;
        }
    }

    /**
     * A bean containing types that can be instantiated from {@link String}s
     */
    private static class StringConversionBean
    {
        Class type;

        byte[] bytes;

        char[] chars;

        Date date;

        public StringConversionBean()
        {
        }

        public StringConversionBean(final Class type)
        {
            this.type = type;
        }

        public Class getType()
        {
            return type;
        }

        public void setType(final Class type)
        {
            this.type = type;
        }

        public byte[] getBytes()
        {
            return bytes;
        }

        public void setBytes(final byte[] bytes)
        {
            this.bytes = bytes;
        }

        public char[] getChars()
        {
            return chars;
        }

        public void setChars(final char[] chars)
        {
            this.chars = chars;
        }

        public Date getDate()
        {
            return date;
        }

        public void setDate(final Date date)
        {
            this.date = date;
        }
    }

    /**
     * A bean containing property setters with multiple arguments
     */
    private static class MultiArgSetterBean
    {
        String foo;

        public String getFoo()
        {
            return foo;
        }

        public void setFoo(final String foo)
        {
            this.foo = foo;
        }

        public void setFoo(final String foo, final String bar)
        {
            this.foo = foo + bar;
        }
    }

    /**
     * A bean containing a child bean.
     */
    private static class ParentBean
    {
        Bean child = new Bean(0, "");

        public Bean getChild()
        {
            return child;
        }

        public void setChild(final Bean child)
        {
            this.child = child;
        }
    }

    /**
     * A bean containing with overloaded setter.
     */
    private static class OverloadedSetterBean
    {
        String string;

        boolean flag;

        public void setProperty(final String string)
        {
            this.string = string;
        }

        public void setProperty(final boolean flag)
        {
            this.flag = flag;
        }

        public String getString()
        {
            return string;
        }

        public boolean getBoolean()
        {
            return flag;
        }
    }

    /**
     * A bean with indexed properties.
     */
    private static class IndexedPropertyBean
    {
        List<String> values = new LinkedList<String>();

        public String getValue(final int index)
        {
            return values.get(index);
        }

        public void setValue(final int index, final String value)
        {
            values.set(index, value);
        }

        public List<String> getValues()
        {
            return values;
        }

        public void setValues(final List<String> values)
        {
            this.values = values;
        }
    }

    /**
     * A bean with keyed properties.
     */
    private static class KeyedPropertyBean
    {
        final Map<String, String> map = new HashMap<String, String>();

        public Map<String, String> getMap()
        {
            return map;
        }
    }
}
