/**
 * Copyright 2005 by jFramework.org.
 */
package org.jframework.core.util;

import junit.framework.TestCase;
import junitx.framework.Assert;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * Abstract Proxy test
 *
 * @author mgridnev
 * @version 1.0, Aug 20, 2005
 */
public class ProxyTest extends TestCase
{
    /**
     * Tests proxy creation
     */
    public void testProxy()
    {
        final String string = "value";
        final Proxy<String> proxy = ProxyUtils.create(string);
        final Proxy<String> proxyOther = new AbstractProxy<String>()
        {
            public String getProxiedObject()
            {
                return string;
            }
        };

        assertEquals(string.hashCode(), proxy.hashCode());
        assertEquals(proxy.hashCode(), proxyOther.hashCode());

        assertEquals(proxy, proxyOther);
        assertFalse(proxy.equals(string)); // only two proxies or two string should compare as equal

        assertEquals(string, proxy.toString());
    }

    /**
     * Tests unproxying
     */
    public void testUnproxy()
    {
        assertNull(ProxyUtils.unproxy((Object) null));

        final String string = "value";
        final Proxy<String> proxy = ProxyUtils.create(string);
        assertSame(string, ProxyUtils.unproxy(proxy));

        final Proxy<Proxy<String>> proxyLevel2 = new AbstractProxy<Proxy<String>>()
        {
            public Proxy<String> getProxiedObject()
            {
                return proxy;
            }
        };
        assertSame(string, ProxyUtils.unproxy(proxyLevel2));
    }

    /**
     * Tests collection proxy
     */
    @SuppressWarnings("unchecked")
    public void testCollectionUnproxy()
    {
        final String one = "one";
        final String two = "two";

        final ArrayList<Proxy<String>> listProxies = new ArrayList<Proxy<String>>();
        listProxies.add(ProxyUtils.create(one));
        listProxies.add(ProxyUtils.create(two));

        // Uproxying the collection effectively unproxies the elements of the collection
        final Collection<String> colUnproxied = (Collection<String>) ProxyUtils.unproxy(listProxies);

        final List<String> listExpected = Arrays.asList(one, two);
        assertEquals(listExpected, colUnproxied);

        assertSame(colUnproxied, ProxyUtils.unproxy(colUnproxied)); // no exception should be thrown
    }

    /**
     * Tests hashcode and equals
     */
    public void testHashCodeAndEquals()
    {
        final String value = "value";
        final String value2 = "value2";

        final Proxy<String> proxy = ProxyUtils.create(value);
        final Proxy<String> proxyClone = ProxyUtils.create(value);
        final Proxy<String> proxyDiff = ProxyUtils.create(value2);
        final Proxy<String> proxyNull = ProxyUtils.create(null);

        assertEquals(proxy, proxy);

        assertEquals(proxy.hashCode(), proxyClone.hashCode());
        assertEquals(proxy, proxyClone);

        Assert.assertNotEquals(proxy.hashCode(), proxyDiff.hashCode());
        Assert.assertNotEquals(proxy, proxyDiff);

        assertEquals(0, proxyNull.hashCode());
        Assert.assertNotEquals(proxy, proxyNull);
    }

    /**
     * Tests string conversion
     */
    public void testToString()
    {
        final String value = "value";
        assertEquals(value, ProxyUtils.create(value).toString());

        assertEquals("null", ProxyUtils.create(null).toString());
    }

    /**
     * Tests comparison
     */
    public void testCompare()
    {
        final Integer one = 1;
        final Integer two = 2;

        final Proxy<Integer> proxyOne = ProxyUtils.create(one);
        final Proxy<Integer> proxyTwo = ProxyUtils.create(two);

        assertEquals(0, proxyOne.compareTo(proxyOne));
        assertEquals(1, proxyTwo.compareTo(proxyOne));
        assertEquals(-1, proxyOne.compareTo(proxyTwo));
    }

    /**
     * Tests simple proxy
     */
    public void testSimpleRroxy()
    {
        final Object value = "value";
        final Proxy proxy = ProxyUtils.create(value);

        assertEquals(value, proxy.getProxiedObject());
    }

    /**
     * Tests collection unproxying
     */
    public void testAsCollection()
    {
        final List list = Arrays.asList("one", "two");

        assertEquals(list, ProxyUtils.unproxy(ProxyUtils.create(list)));
    }
}
