/**
 * Copyright 2004 by jFramework.org.
 */
package org.jframework.core.util;

import junit.framework.TestCase;

import java.util.regex.Pattern;

/**
 * Tests regular expressions
 *
 * @author mgridnev
 * @version 1.0, Jan 21, 2005
 */
public class RegexTest extends TestCase
{
    // NOTE: Relaxed Not Empty is meaningless

    public void testStrictNotEmpty()
    {
        final Pattern pattern = Regex.Strict.notEmpty();

        assertRegexTrue("a", pattern);
        assertRegexTrue("abc ", pattern);
        assertRegexTrue(" abc ", pattern);
        assertRegexTrue(" abc abc ", pattern);

        assertRegexFalse("", pattern);
        assertRegexFalse(" ", pattern);
        assertRegexFalse("    ", pattern);
    }

    public void testRelaxedText()
    {
        final Pattern pattern = Regex.Relaxed.text(5);

        assertRegexTrue("", pattern);
        assertRegexTrue("4 .&^", pattern);
        assertRegexTrue("1\n2", pattern);

        assertRegexFalse("123456", pattern);
    }

    public void testStrictText()
    {
        // No min
        Pattern pattern = Regex.Strict.text(0, 3);

        assertRegexTrue("", pattern);
        assertRegexTrue("1a1", pattern);

        assertRegexFalse("1a1b", pattern);

        // Min max
        pattern = Regex.Strict.text(2, 4);

        assertRegexTrue("1a", pattern);
        assertRegexTrue("1a b", pattern);

        assertRegexFalse("1", pattern);
        assertRegexFalse("1a1 b", pattern);
    }

    public void testRelaxedDigits()
    {
        final Pattern pattern = Regex.Relaxed.digits(3);

        assertRegexTrue("", pattern);
        assertRegexTrue("1", pattern);
        assertRegexTrue("123", pattern);

        assertRegexFalse(" ", pattern);
        assertRegexFalse("-1", pattern);
        assertRegexFalse("a", pattern);

        assertRegexTrue("1234", Regex.Relaxed.digits(-1));
    }

    public void testStrictDigits()
    {
        final Pattern pattern = Regex.Strict.digits(1, 2);

        assertRegexFalse("", pattern);
        assertRegexTrue("1", pattern);
        assertRegexTrue("12", pattern);
        assertRegexFalse("123", pattern);

        assertRegexFalse(" ", pattern);
        assertRegexFalse("-1", pattern);
        assertRegexFalse("a", pattern);
    }

    public void testRelaxedHexDigits()
    {
        final Pattern pattern = Regex.Relaxed.hex(3);

        assertRegexTrue("", pattern);
        assertRegexTrue("1", pattern);
        assertRegexTrue("123", pattern);
        assertRegexTrue("1De", pattern);

        assertRegexFalse(" ", pattern);
        assertRegexFalse("-1", pattern);
        assertRegexFalse("x", pattern);
        assertRegexFalse("1g", pattern);

        assertRegexTrue("1234a", Regex.Relaxed.hex(-1));
    }

    public void testStrictHexDigits()
    {
        final Pattern pattern = Regex.Strict.hex(1, 2);

        assertRegexFalse("", pattern);
        assertRegexTrue("1", pattern);
        assertRegexTrue("1a", pattern);
        assertRegexTrue("1D", pattern);
        assertRegexTrue("D1", pattern);
        assertRegexFalse("111", pattern);

        assertRegexFalse(" ", pattern);
        assertRegexFalse("-1", pattern);
        assertRegexFalse("x", pattern);
        assertRegexFalse("1g", pattern);
    }

    public void testRelaxedIP()
    {
        final Pattern pattern = Regex.Relaxed.IP();

        assertRegexTrue("", pattern);
        assertRegexTrue("1", pattern);
        assertRegexTrue("1.", pattern);
        assertRegexTrue("111", pattern);

        assertRegexFalse("a", pattern);
        assertRegexFalse(" ", pattern);
        assertRegexFalse("111.111.111.111.", pattern);
    }

    public void strictIP()
    {
        final Pattern pattern = Regex.Relaxed.IP();

        assertRegexTrue("0.0.0.0", pattern);
        assertRegexTrue("1.1.1.1", pattern);
        assertRegexTrue("111.111.111.111", pattern);
        assertRegexTrue("0.0.0.0.", pattern);

        assertRegexFalse("111.111.111.111.", pattern);
    }

    // NOTE: Relax Netmask is the same as relaxed IP

    public void testStrictNetmask()
    {
        final Pattern pattern = Regex.Strict.netMask();

        final String octets[] = {"0", "128", "192", "224", "240", "248", "252", "254", "255"}; // valid octets
        final int cParts = 4; // 4 parts of the netmask

        // Test valid
        for (int iLeading = 0; iLeading < cParts; iLeading++) {
            final StringBuffer sbLeading = new StringBuffer();

            for (int i = 0; i < iLeading; i++) {
                sbLeading.append(octets[octets.length - 1]);
                sbLeading.append(".");
            }

            for (String rgOctet : octets) {
                final StringBuffer sbFinal = new StringBuffer(sbLeading.toString());

                sbFinal.append(rgOctet);

                for (int iTrailing = 0; iTrailing < cParts - iLeading - 1; iTrailing++) {
                    sbFinal.append(".");
                    sbFinal.append(octets[0]);
                }

                assertRegexTrue(sbFinal.toString(), pattern);
            }
        }

        // Test invalid
        final String[] invalidMasks = {"128", "128.0", "128.0.0", "128.128.0.0", "255.255.224.255", "0.255.255.255"};
        for (String mask : invalidMasks) {
            assertRegexFalse(mask, pattern);
        }
    }

    public void testRelaxedHostName()
    {
        final Pattern pattern = Regex.Relaxed.hostName();

        assertRegexFalse(" ", pattern);
        assertRegexFalse(".", pattern);
        assertRegexTrue("aB1-", pattern); // only letters digits and dash

        assertMinMaxLength(0, 63, "a", pattern); // up to 63 chars
    }

    public void testStrictHostName()
    {
        final Pattern pattern = Regex.Strict.hostName();

        assertRegexFalse(" ", pattern);
        assertRegexTrue("a", pattern);
        assertRegexFalse("1", pattern);
        assertRegexTrue("a1", pattern);
        assertRegexFalse("a-", pattern);
        assertRegexTrue("a-B", pattern);
        assertRegexTrue("B-1", pattern);
        assertRegexFalse("a.b", pattern);
        assertRegexFalse("a b", pattern);

        assertMinMaxLength(1, 63, "a", pattern);
    }

    public void testRelaxedDomain()
    {
        final Pattern pattern = Regex.Relaxed.domain();

        assertRegexTrue("www.domain.com", pattern);
        assertMinMaxLength(0, 128, "a", pattern);
    }

    public void testStrictDomain()
    {
        final Pattern pattern = Regex.Strict.domain();

        assertRegexTrue("domain.com", pattern);
        assertRegexTrue("3w.domain.com", pattern);
        assertRegexTrue("www.domain.com", pattern);

        assertRegexFalse("1.2.3.4", pattern);
        assertRegexFalse("www.123", pattern);
        assertRegexFalse("www.a b.com", pattern);
    }

    public void testRelaxedHost()
    {
        final Pattern pattern = Regex.Relaxed.host();

        assertRegexTrue("", pattern);
        assertRegexFalse(" ", pattern);
        assertRegexTrue("a1A-.", pattern);

        assertMinMaxLength(0, 128, "a", pattern);
    }

    public void testStrictHost()
    {
        final Pattern pattern = Regex.Strict.host();

        assertRegexTrue("a", pattern);
        assertRegexFalse("1", pattern);
        assertRegexFalse("a-", pattern);
        assertRegexTrue("1.2.3.4", pattern);
        assertRegexTrue("a.b.com", pattern);

        assertMinMaxLength(1, 63, "a", pattern);

        assertRegexFalse("", pattern);
        assertRegexFalse("a.b", pattern);
        assertRegexFalse("a_b", pattern);
        assertRegexFalse("a b", pattern);
    }

    public void testRelaxedPort()
    {
        final Pattern pattern = Regex.Relaxed.port();

        assertRegexTrue("", pattern);
        assertRegexTrue("1", pattern);
        assertRegexTrue("99999", pattern);

        assertRegexFalse("a", pattern);
        assertRegexFalse("999999", pattern);
    }

    public void testStrictPort()
    {
        final Pattern pattern = Regex.Strict.port();

        assertRegexTrue("1", pattern);
        assertRegexTrue("99999", pattern);

        assertRegexFalse("", pattern);
        assertRegexFalse("a", pattern);
        assertRegexFalse("999999", pattern); // not very accurate
    }

    public void testRelaxedMD5()
    {
        final Pattern pattern = Regex.Relaxed.MD5();

        assertMinMaxLength(0, 32, "a", pattern);
    }

    public void testStrictMD5()
    {
        final Pattern pattern = Regex.Strict.MD5();

        assertMinMaxLength(32, 32, "a", pattern);
    }

    public void testStrictURL()
    {
        final Pattern pattern = Regex.Strict.URL(new String[] {"http", "https"});

        assertRegexTrue("http://www.domain.com", pattern);
        assertRegexTrue("https://www.domain.com", pattern);

        assertRegexTrue("http://172.0.0.1", pattern);
        assertRegexTrue("http://host", pattern);

        assertRegexTrue("http://www.domain.com:80", pattern);

        assertRegexTrue("http://user@www.domain.com:80", pattern);
        assertRegexTrue("http://user:password@www.domain.com:80", pattern);

        assertRegexTrue("http://www.domain.com/dir/subdir/file.html", pattern);
        assertRegexTrue("http://www.domain.com/dir/sub dir/file name.html", pattern); // space not allowed in RFC but we have to allow it

        assertRegexTrue("http://www.domain.com?a=b&c=d", pattern);
        assertRegexTrue("http://www.domain.com/dir?a=b&c=d", pattern);

        assertRegexFalse("http://", pattern);
        assertRegexFalse("www.domain.com", pattern);
    }

    private static void assertRegexTrue(String match, Pattern pattern)
    {
        assertRegex("assertTrue", match, pattern, true);
    }

    private static void assertRegexFalse(String match, Pattern pattern)
    {
        assertRegex("assertFalse", match, pattern, false);
    }

    private void assertMinMaxLength(int min, int max, String template, Pattern pattern)
    {
        final char[] rgInput = new char[max + 1];
        final char[] rgTemplate = template.toCharArray();

        // Fill the input buffer with template
        for (int off = 0; off < rgInput.length; off += rgTemplate.length) {
            System.arraycopy(rgTemplate, 0, rgInput, off, Math.min(rgTemplate.length, rgInput.length - off));
        }

        // One under
        if (min > 0) {
            assertRegexFalse(new String(rgInput, 0, min - 1), pattern);
        }

        // Exact min
        assertRegexTrue(new String(rgInput, 0, min), pattern);

        // Exact max
        assertRegexTrue(new String(rgInput, 0, max), pattern);

        // One over
        assertRegexFalse(new String(rgInput), pattern);
    }

    private static void assertRegex(String message, String match, Pattern pattern, boolean matches)
    {
        final boolean fResult = pattern.matcher(match).matches();
        assertTrue(message + ": '" + match + "'", fResult == matches);
    }
}
